import os
import datetime
import json
import logging
import shutil
from subprocess import Popen, PIPE
from threading import Thread

# {CATTURA_DIR}/bin/apis/python
CATTURA_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
CATTURA_LOG_DIR = os.path.join(CATTURA_DIR,'logs')

loggers = {}
LOG_DATE_FORMAT = "%Y-%m-%d.%I-%M-%S-%p"
LOG_RECORD_FORMAT = "%(asctime)s [%(name)s] %(levelname)s: %(message)s"

GRAPH_JSON = "graph.json"
PLUGIN_JSON = "plugin.json"
OUTPUT_JSON = "output.json"

USER_SETTINGS = "public_settings.json"

# Logging
def getLogger (name):
    if not name in loggers:
        logger = logging.getLogger(name)
        logdir = os.path.join(CATTURA_LOG_DIR, name)
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        logfilename = os.path.join(logdir,"{date}.log".format(date=datetime.datetime.today().strftime(LOG_DATE_FORMAT)))
        
        # handlers
        logFileHandler = logging.FileHandler(logfilename)
        logStreamHandler = logging.StreamHandler()

        # formatter
        logFormatter = logging.Formatter(LOG_RECORD_FORMAT)
        logFileHandler.setFormatter(logFormatter)
        logStreamHandler.setFormatter(logFormatter)

        logger.addHandler(logStreamHandler)
        logger.addHandler(logFileHandler)
        loggers[name] = logger
    return loggers[name]

# Processing Agent
class MediaProcessingAgent ():
    def __init__(self,path=None,quiet=False):
        if path is None:
            path = "media-processing-agent-1.0.0.jar"
        self.path = path
        self.quiet = quiet
        self.process = None
        self.listeners = []

    def on (self,event):
        def register (func):
            if hasattr(event,'__call__'):
                def wrapped (notification):
                    if event(notification):
                        func(notification)
            else:
                def wrapped (notification):
                    if notification['notification'] == event:
                        func(notification)
            self.listeners.append(wrapped)
        return register

    def onMessage (self,notification):
        for listener in self.listeners:
            listener(notification)

    def sendNotification (self,name,body=None):
        if self.process is None:
            return

        self.process.stdin.write(json.dumps({
            "notification": name,
            "body": body
        }))
        self.process.stdin.write("\n")
        self.process.stdin.flush()

    def start (self):
        self.process = Popen(['java','-jar',self.path],stdin=PIPE,stdout=PIPE,stderr=PIPE if self.quiet else None)
        def listener():
            while True:
                try:
                    line = self.process.stdout.readline()
                    try:
                        note = json.loads(line)
                        self.onMessage(note)
                    except ValueError:
                        if not self.quiet:
                            print(line)
                except KeyboardInterrupt: break
        Thread(target=listener).start()

    def kill (self):
        if self.process is None:
            return

        self.process.kill()

# Templates
class Template ():
    def __init__ (self, path):
        self.path = path
        self.id = os.path.basename(path)
        self.graph = Graph(self.path)

    def serialize (self):
        return self.graph.serialize()

    def nodes (self):
        return self.graph.nodes()

    def save (self):
        self.graph.save()

# Graph
class Graph ():
    def __init__ (self, path):
        self.path = path
        self.graphPath = os.path.join(self.path, GRAPH_JSON) if os.path.isdir(self.path) else self.path
        with open(self.graphPath) as graph:
            self.graph = json.loads(graph.read())
            for node in self.graph['nodes']:
                if isinstance(node['config'], str):
                    node['config'] = json.loads(node['config'])

    def serialize (self):
        nodes = []
        for node in self.graph['nodes']:
            newNode = {}
            newNode.update(node)
            if not isinstance(newNode['config'], str):
                newNode['config'] = json.dumps(newNode['config'])
            nodes.append(newNode)
        return json.dumps({"nodes": nodes})

    def nodes (self):
        for node in self.graph['nodes']:
            yield Node(self, node)

    def remove (self, node):
        for raw_node in self.graph['nodes']:
            if raw_node['nodeID'] == node.nodeID:
                self.graph['nodes'].remove(raw_node)

    def save (self):
        with open(self.graphPath, 'w') as graph:
            graph.write(self.serialize())


class Node(object):
    def __init__ (self, template, node):
        object.__setattr__(self, "node", node)
        object.__setattr__(self, "template", template)
    
    def __getattr__ (self, name):
        return self.node[name]
    
    def __setattr__ (self, name, value):
        self.node[name] = value

def templates ():
    templateDir = os.path.join(CATTURA_DIR, "storage", "templates")
    for item in os.listdir(templateDir):
        try:
            yield Template(os.path.join(templateDir, item))
        except Exception as e: print(e)

# Plugins
class Plugin ():
    def __init__(self, pluginType, path):
        self.path = path
        self.id = os.path.basename(path)
        self.pluginType = pluginType
        self.configDirectory = os.path.join(path, "conf")

        self.data = {}
        try:
            with open(os.path.join(path, PLUGIN_JSON)) as pluginJson:
                self.data.update(json.loads(pluginJson.read()))
        except Exception as e: print(e)

    def config (self):
        config = {}
        try:
            for item in os.listdir(self.configDirectory):
                with open(os.path.join(self.configDirectory, item)) as configFile:
                    config.update(json.loads(configFile.read()))
        except Exception as e: print(e)
        return config

    def setConfig (self, config):
        configFileName = "configuration.conf"

        try:
            os.mkdir(self.configDirectory)
        except Exception: pass
        
        for item in os.listdir(self.configDirectory):
            os.remove(os.path.join(self.configDirectory, item))

        with open(os.path.join(self.configDirectory, configFileName), "w") as configFile:
            configFile.write(json.dumps(config))

    def wipe (self):
        shutil.rmtree(self.path, ignore_errors=True)

def plugins ():
    pluginDir = os.path.join(CATTURA_DIR, "plugins")
    for pluginType in os.listdir(pluginDir):
        for item in os.listdir(os.path.join(pluginDir, pluginType)):
            try:
                yield Plugin(pluginType, os.path.join(pluginDir, pluginType, item))
            except Exception as e: print(e)

# Settings
def userSettings ():
    confDirectory = os.path.join(CATTURA_DIR, "conf")
    with open(os.path.join(confDirectory, USER_SETTINGS)) as userSettings:
        return json.loads(userSettings.read())

def setUserSettings (settings):
    confDirectory = os.path.join(CATTURA_DIR, "conf")
    with open(os.path.join(confDirectory, USER_SETTINGS), "w") as userSettings:
        userSettings.write(json.dumps(settings))

# Recordings
class Recording ():
    def __init__ (self, path):
        self.path = path
        self.id = os.path.basename(path)
        self.graph = Graph(path)

    def outputs (self):
        for node in self.graph.nodes():
            outputJsonPath = os.path.join(node.workingDir, OUTPUT_JSON)
            if not os.path.isfile(outputJsonPath):
                continue
            
            with open(outputJsonPath) as outputJson:
                for key, value in json.loads(outputJson.read()).items():
                    if "path" in value and not os.path.exists(value["path"]):
                        value["path"] = os.path.join(node.workingDir, value["path"])
                    yield (key, value)
        
def recordings ():
    recordingsDir = os.path.join(CATTURA_DIR, "storage", "capture_sessions")
    for (recordingDir, subdirectories, filenames) in os.walk(recordingsDir):
        if not GRAPH_JSON in filenames:
            continue
        
        try:
            yield Recording(os.path.join(recordingsDir, recordingDir))
        except Exception as e: print(e)